--
-- Please see the license.html file included with this distribution for
-- attribution and copyright information.
--

local aTraitRolls = {}

function onInit()

	-- register pre-defined trait rolls
	registerTraitRoll("shakenrecovery", shakenRecoveryRollHandler)
	registerTraitRoll("soak", soakRollHandler)
	registerTraitRoll("incapacitation", incapacitationRollHandler)
end

function registerTraitRoll(sTraitType, fHandler)
	local sKey = StringManager.simplify(sTraitType)
	aTraitRolls[sKey] = fHandler
end

function unregisterTraitRoll(sTraitType)
	local sKey = StringManager.simplify(sTraitType)
	aTraitRolls[sKey] = nil
end

function makeTraitRoll(sActorType, nodeActor, sTraitType, rUserData)
	for sKey, fTraitHandler in pairs(aTraitRolls) do
		if sKey == StringManager.simplify(sTraitType) then
			ModifierManager.applyTraitModifiers(sActorType, nodeActor, sTraitType)
			fTraitHandler(sActorType, nodeActor, sTraitType, rUserData)
		end
	end
end

function makeTraitDrag(sActorType, nodeActor, sTraitType, draginfo, rUserData)
	for sKey, fTraitHandler in pairs(aTraitRolls) do
		if sKey == StringManager.simplify(sTraitType) then
			ModifierManager.applyTraitModifiers(sActorType, nodeActor, sTraitType)
			fTraitHandler(sActorType, nodeActor, sTraitType, rUserData, draginfo)
		end
	end
end

--
-- Make attribute rolls
--

function makeAttributeRoll(sActorType, nodeActor, sAttribute, rUserData)
	local nodeTrait = getTraitNode(sActorType, nodeActor, StringManager.simplify(sAttribute))
	rollTrait(CharacterManager.getActorShortcut(sActorType, nodeActor), nodeTrait, rUserData)
end

function makeAttributeDrag(draginfo, sActorType, nodeActor, sAttribute, rUserData)
	local nodeTrait = getTraitNode(sActorType, nodeActor, StringManager.simplify(sAttribute))
	return traitRollDrag(CharacterManager.getActorShortcut(sActorType, nodeActor), nodeTrait, rUserData, draginfo)
end

--
-- Make skill rolls
--

function makeSkillRoll(sActorType, nodeActor, sSkillName, rUserData)
	local _, nodeChar = asCharacterType(sActorType, nodeActor)
	local nodeTrait = SkillManager.getSkillNode(nodeChar, sSkillName)
	rollTrait(CharacterManager.getActorShortcut(sActorType, nodeActor), nodeTrait, rUserData)
end

function makeSkillDrag(draginfo, sActorType, nodeActor, sSkillName, rUserData)
	local _, nodeChar = asCharacterType(sActorType, nodeActor)
	local nodeTrait = SkillManager.getSkillNode(nodeChar, sSkillName)
	return traitRollDrag(CharacterManager.getActorShortcut(sActorType, nodeActor), nodeTrait, rUserData, draginfo)
end

--
-- Pre-defined trait rolls
--

function shakenRecoveryRollHandler(sActorType, nodeActor, sTraitType, rUserData, draginfo)
	ModifierStack.applyEffectModifierOnEntity(sActorType, nodeActor, "shaken")
	local sDescPrefix = "Shaken Recovery"
	local nodeTrait = getTraitNode(sActorType, nodeActor, "spirit")
	if draginfo then
		traitRollDrag(CharacterManager.getActorShortcut(sActorType, nodeActor), nodeTrait, rUserData, draginfo, sDescPrefix, sTraitType)
	else
		rollTrait(CharacterManager.getActorShortcut(sActorType, nodeActor), nodeTrait, rUserData, sDescPrefix, sTraitType)
	end
end

function soakRollHandler(sActorType, nodeActor, sTraitType, rUserData, draginfo)
	ModifierStack.applyEffectModifierOnEntity(sActorType, nodeActor, "soak")
	local sDescPrefix = "Soak"
	local nodeTrait = getTraitNode(sActorType, nodeActor, "vigor")
	if draginfo then
		traitRollDrag(CharacterManager.getActorShortcut(sActorType, nodeActor), nodeTrait, rUserData, draginfo, sDescPrefix, sTraitType)
	else
		rollTrait(CharacterManager.getActorShortcut(sActorType, nodeActor), nodeTrait, rUserData, sDescPrefix, sTraitType)
	end
end

function incapacitationRollHandler(sActorType, nodeActor, sTraitType, rUserData, draginfo)
	ModifierStack.applyEffectModifierOnEntity(sActorType, nodeActor, "incapacitation")
	local sDescPrefix = "Incapacitation"
	local nodeTrait = getTraitNode(sActorType, nodeActor, "vigor")
	if draginfo then
		traitRollDrag(CharacterManager.getActorShortcut(sActorType, nodeActor), nodeTrait, rUserData, draginfo, sDescPrefix, sTraitType)
	else
		rollTrait(CharacterManager.getActorShortcut(sActorType, nodeActor), nodeTrait, rUserData, sDescPrefix, sTraitType)
	end
end

--
-- Roll builders
--

function getTraitRoll(rActor, nodeTrait, sTraitType, rUserData, sDescPrefix)
	local sActorType, nodeActor = CharacterManager.getActor(rActor)
	local nTotalModifier, sDescription = getTotalTraitModifier(sActorType, nodeActor, nodeTrait)
	if sDescPrefix and sDescPrefix ~= "" then
		sDescription = sDescPrefix .. ": " .. sDescription
	end
	local aDices, aTraitDice = getTraitDice(sActorType, nodeActor, nodeTrait)
	return { type = "traitdice",
		 description = sDescription,
		 modifier = nTotalModifier,
		 dices = aDices,
		 traitDice = aTraitDice,
		 actor = CharacterManager.getActorShortcut(sActorType, nodeActor),
		 custom = getTraitCustomData(sActorType, nodeActor, nodeTrait, sTraitType, rUserData) }
end

function getAttackRoll(rActor, nodeAttack, sTraitType, rUserData)
	local sActorType, nodeActor = CharacterManager.getActor(rActor)
	local _, nodeChar = asCharacterType(sActorType, nodeActor)
	local nodeTrait = SkillManager.getSkillNode(nodeChar, getAssociatedSkill(nodeChar, nodeAttack))
	local nTotalModifier, sDescription = getTotalAttackModifier(sActorType, nodeActor, nodeTrait, nodeAttack)
	local aDices, aTraitDice = getTraitDice(sActorType, nodeActor, nodeTrait, nodeAttack)
	return { type = "attackdice",
		 description = sDescription,
		 modifier = nTotalModifier,
		 dices = aDices,
		 traitDice = aTraitDice,
		 actor = CharacterManager.getActorShortcut(sActorType, nodeActor),
		 traitNode = nodeTrait,
		 custom = getAttackCustomData(sActorType, nodeActor, nodeAttack, sTraitType, rUserData) }
end

function getDamageRoll(rActor, nodeAttack, sTraitType, rUserData)
	local sActorType, nodeActor = CharacterManager.getActor(rActor)
	return { type = "damagedice",
		 description = getDamageDescription(sActorType, nodeActor, nodeAttack),
		 modifier = getDamageModifier(sActorType, nodeActor, nodeAttack),
		 dices = getDamageDice(sActorType, nodeActor, nodeAttack),
		 actor = CharacterManager.getActorShortcut(sActorType, nodeActor),
		 custom = getDamageCustomData(sActorType, nodeActor, nodeAttack, sTraitType, rUserData) }
end

--
-- Roll markers
--

function traitRollDrag(rActor, nodeTrait, rUserData, draginfo, sDescPrefix, sTraitType)
	return rollDrag(getTraitRoll(rActor, nodeTrait, sTraitType, rUserData, sDescPrefix), nodeTrait, sTraitType, draginfo, rUserData)
end

function rollTrait(rActor, nodeTrait, rUserData, sDescPrefix, sTraitType)
	roll(getTraitRoll(rActor, nodeTrait, sTraitType, rUserData, sDescPrefix), nodeTrait, sTraitType, rUserData)
end

function attackRollDrag(rActor, nodeAttack, rUserData, draginfo, sTraitType)
	local rRoll = getAttackRoll(rActor, nodeAttack, sTraitType, rUserData)
	return rollDrag(rRoll, rRoll.traitNode, sTraitType, draginfo, rUserData)
end

function rollAttack(rActor, nodeAttack, rUserData, sTraitType)
	local rRoll = getAttackRoll(rActor, nodeAttack, sTraitType, rUserData)
	roll(rRoll, rRoll.traitNode, sTraitType, rUserData)
end

function damageRollDrag(rActor, nodeAttack, rUserData, draginfo, sTraitType)
	return rollDrag(getDamageRoll(rActor, nodeAttack, sTraitType, rUserData), nodeAttack, sTraitType, draginfo, rUserData)
end

function rollDamage(rActor, nodeAttack, rUserData, sTraitType)
	roll(getDamageRoll(rActor, nodeAttack, sTraitType, rUserData), nodeAttack, sTraitType, rUserData)
end

function rollDrag(rRoll, nodeTrait, sTraitType, draginfo, rUserData)
	draginfo.setType(rRoll.type)
	draginfo.setDieList(rRoll.dices)
	draginfo.setDescription(rRoll.description)
	draginfo.setStringData(rRoll.description)
	draginfo.setNumberData(rRoll.modifier)

	DragDataUtil.encodeActors(draginfo, rRoll.actor)
	DragDataUtil.convertTableToMetaData(draginfo, rRoll.custom)
	DragDataUtil.addSubType(draginfo, sTraitType)

	applyModifiers(rRoll, nodeTrait, sTraitType, rUserData)

	return draginfo
end

function roll(rRoll, nodeTrait, sTraitType, rUserData)
	applyModifiers(rRoll, nodeTrait, sTraitType, rUserData)
	RollsManager.throwDice(rRoll.type, rRoll.dices, rRoll.modifier, rRoll.description, rRoll.custom, rRoll.actor)
end

function applyModifiers(rRoll, nodeTrait, sTraitType, rUserData)
	if StringManager.contains({"traitdice", "attackdice"}, rRoll.type) then
		RollsManager.applyTraitRollModifiers(rRoll.actor, rRoll.description, nodeTrait, sTraitType)
	elseif StringManager.contains({"damagedice"}, rRoll.type) then
		RollsManager.applyDamageRollModifiers(rRoll.actor, rRoll.description, nodeTrait, sTraitType)
	end
	if rUserData then
		if rUserData.ignoreJokerBonus == 1 then
			ModifierStack.removeByName("Joker Bonus")
		end
	end
end

--
-- Description info
--

function getDescription(nodeTrait)
	if not nodeTrait then
		return ""
	elseif isAttributeNode(nodeTrait) then
		return StringManager.capitalize(nodeTrait.getName())
	else
		return DB.getValue(nodeTrait.getParent(), "name", "")
	end
end

function getAttackDescription(nodeAttack)
	local sDesc = ""
	if nodeAttack then
		sDesc = DB.getValue(nodeAttack, "name", "")
		if DB.getValue(nodeAttack, "masternode", "") ~= "" then
			local sSubName = DB.getValue(nodeAttack, "subname", "")
			if #sSubName > 0 then
				sDesc = sDesc .. " (" .. sSubName .. ")"
			end
		end
	end
	return sDesc
end

function getDamageDescription(sActorType, nodeActor, nodeAttack)
	local sDesc = getAttackDescription(nodeAttack)
	if #sDesc > 0 then
		sDesc = sDesc .. " "
	end
	sDesc = sDesc .. "Damage"
	local rAttack = AttackManager.getAttack(sActorType, nodeActor, nodeAttack)
	if rAttack.damageAffectedEffects then
		sDesc = sDesc .. " (Affected by " .. rAttack.damageAffectedEffects ..")"
	end
	if nodeAttack and DB.getValue(nodeAttack, "damagetype", "") == "nonlethal" then
		sDesc = sDesc .. " (Nonlethal)"
	end
	return sDesc
end

--
-- Trait die info
--

function getTraitDice(sActorType, nodeActor, nodeTrait, nodeAttack)
	if not nodeTrait or nodeTrait.getType() ~= "dice" then
		return {}
	else
		local aTraitDice = nodeTrait.getValue() or {}
		local rTraitEffect = EffectManager.getTraitAdjustmentEffect(sActorType, nodeActor, nodeTrait)
		if rTraitEffect then
			aTraitDice = SWDiceManager.getAdjustedDice(nodeTrait, rTraitEffect)
		end

		local aDices = {}
		for _, sDice in pairs(aTraitDice) do
			table.insert(aDices, sDice)
		end

		local rActor = CharacterManager.getActorShortcut(sActorType, nodeActor)
		if hasWildDie(rActor) then
			table.insert(aDices, getWildDie(rActor, nodeTrait, nodeAttack))
		end

		return aDices, aTraitDice
	end
end

function getDamageDice(sActorType, nodeActor, nodeAttack)
	if not nodeAttack then
		return {}
	else
		local rAttack = AttackManager.getAttack(sActorType, nodeActor, nodeAttack)
		return rAttack.damagedice or DB.getValue(nodeAttack, "damagedice", {})
	end
end

--
-- Wild die info
--

function hasWildDie(rActor)
	local _, nodeChar = asCharacterType(rActor)
	if nodeChar then
		return DB.getValue(nodeChar, "wildcard", 0) == 1
	end
end

function getWildDie(rActor, nodeTrait, nodeAttack)
	local sWildDie = getAttackWildDie(nodeAttack)
	if sWildDie == "" then
		sWildDie = getTraitWildDie(nodeTrait)
		if sWildDie == "" then
			local _, nodeChar = asCharacterType(rActor)
			sWildDie = getCharacterWildDie(nodeChar)
		end
	end
	return sWildDie
end

function getAttackWildDie(nodeAttack)
	if not nodeAttack then
		return ""
	else
		return DB.getValue(nodeAttack, "wilddie", "")
	end
end

function getTraitWildDie(nodeTrait)
	if not nodeTrait then
		return ""
	elseif isAttributeNode(nodeTrait) then
		return DB.getValue(nodeTrait.getParent(), nodeTrait.getName() .. "WildDie", "")
	else
		return DB.getValue(nodeTrait.getParent(), "wilddie", "")
	end
end

function getCharacterWildDie(nodeChar)
	local sWildDie = getDefaultWildDie()
	if nodeChar then
		sWildDie = DB.getValue(nodeChar, "wilddie", "")
		if sWildDie == "" then
			sWildDie = getDefaultWildDie()
		end
	end
	return sWildDie
end

function getDefaultWildDie()
	return CampaignRegistry.wilddietype ~= "" and CampaignRegistry.wilddietype or "d6"
end

--
-- Modifier info
--

function getTotalModifier(vActor, vSource, nodeTrait, nodeAttack, bApplyEffectModifiers)

	local sActorType = vActor
	local nodeActor = vSource
	if type(vActor) == "table" then
		sActorType, nodeActor = CharacterManager.getActor(vActor)
		nodeTrait = vSource
	end

	local nModifier = getTraitModifier(nodeTrait)
	local sDescription = getDescription(nodeTrait)
	local aAffectedBy = {}

	local rTraitEffect = EffectManager.getTraitAdjustmentEffect(sActorType, nodeActor, nodeTrait)
	if bApplyEffectModifiers and rTraitEffect then
		_, nModifier = SWDiceManager.getAdjustedDice(nodeTrait, rTraitEffect)
		table.insert(aAffectedBy, rTraitEffect.shortlabel)
	end

	local nBonusModifier, sBonusDesc = getTraitBonusListTotalModifier(nodeTrait)
	local nTotalModifier = nModifier + nBonusModifier

	local nAttackBonusModifier, sAttackBonusDesc = getAttackBonusListTotalModifier(nodeAttack)
	if sAttackBonusDesc ~= "" then

		-- Adjust ranged MinStr penalty
		local sEffectAffectingMinStr = AttackManager.getEffectsAffectingRangedMinStrRollModifier(sActorType, nodeActor, nodeAttack, sAttackBonusDesc)
		if sEffectAffectingMinStr then
			table.insert(aAffectedBy, sEffectAffectingMinStr)
		end
		nAttackBonusModifier, sAttackBonusDesc = AttackManager.adjustRangedMinStrRollModifier(sActorType, nodeActor, nodeAttack, nAttackBonusModifier, sAttackBonusDesc)
		nTotalModifier = nTotalModifier + nAttackBonusModifier
		if sAttackBonusDesc ~= "" then
			if sBonusDesc ~= "" then
				sBonusDesc = sBonusDesc .. ", "
			end
			sBonusDesc = sBonusDesc .. sAttackBonusDesc
		end
	end

	local sAffectedBy = StringManager.join(aAffectedBy, ", ")
	if sAffectedBy then
		sDescription = sDescription .." (Affected by " .. sAffectedBy .. ")"
	end

	if nModifier ~= 0 then
		if nModifier > 0 then
			sDescription = sDescription .. " +" .. nModifier
		else
			sDescription = sDescription .. " " .. nModifier
		end
	end

	if sBonusDesc ~= "" then
		sDescription = sDescription .. " {" .. sBonusDesc .. "}"
	end

	return nTotalModifier, sDescription
end

function getTotalTraitModifier(sActorType, nodeActor, nodeTrait)
	local nTotalModifier, sDescription = getTotalModifier(sActorType, nodeActor, nodeTrait, nil, true)
	return nTotalModifier, sDescription
end

function getTotalAttackModifier(sActorType, nodeActor, nodeTrait, nodeAttack)

	local sTraitName = getAssociatedSkill(nodeActor, nodeAttack)
	local nTotalModifier, sDescription = getTotalModifier(sActorType, nodeActor, nodeTrait, nodeAttack, true)

	local sTraitDescription = sDescription
	local _, nodeChar = asCharacterType(sActorType, nodeActor)
	if not SkillManager.getSkill(nodeChar, sTraitName) then
		if sTraitName ~= "" then
			sDescription = sTraitName .. " (Unskilled)"
		else
			sDescription = "Unskilled"
		end
	end

	local sAttackDescription = getAttackDescription(nodeAttack)
	if StringManager.simplify(sTraitDescription) ~= StringManager.simplify(sAttackDescription) then
		if sTraitDescription ~= "" then
			sDescription = sDescription .. ": "
		end
		sDescription = sDescription .. sAttackDescription
	end

	return nTotalModifier, sDescription
end

function getTraitModifier(nodeTrait)
	if not nodeTrait then
		return 0
	elseif isAttributeNode(nodeTrait) then
		return DB.getValue(nodeTrait.getParent(), nodeTrait.getName() .. "Mod", 0)
	else
		return DB.getValue(nodeTrait.getParent(), "skillmod", 0)
	end
end

function getDamageModifier(sActorType, nodeActor, nodeAttack)
	if not nodeAttack then
		return 0
	else
		local rAttack = AttackManager.getAttack(sActorType, nodeActor, nodeAttack)
		return rAttack.damagebonus or DB.getValue(nodeAttack, "damagebonus", 0)
	end
end

function getTraitBonusListTotalModifier(nodeTrait)
	local nodeBonusList = getTraitBonusListNode(nodeTrait)
	local nBonusModifier, sBonusDesc = getBonusListTotalModifier(nodeBonusList)
	return nBonusModifier, sBonusDesc
end

function getAttackBonusListTotalModifier(nodeAttack)
	local nodeBonusList = getAttackBonusListNode(nodeAttack)
	local nBonusModifier, sBonusDesc = getBonusListTotalModifier(nodeBonusList)
	return nBonusModifier, sBonusDesc
end

function getTraitBonusListNode(nodeTrait)
	if nodeTrait then
		local nodeBonusList = nodeTrait.getParent().createChild("bonuslist")
		if nodeBonusList and isAttributeNode(nodeTrait) then
			nodeBonusList = nodeBonusList.createChild(nodeTrait.getName())
		end
		return nodeBonusList
	end
end

function getAttackBonusListNode(nodeAttack)
	if nodeAttack then
		return nodeAttack.createChild("bonuslist")
	end
end

function getBonusListTotalModifier(nodeBonusList)
	local nBonusModifier = 0
	local sBonusDesc = ""
	if nodeBonusList then
		nBonusModifier, sBonusDesc = BonusList.new(nodeBonusList).getTotal()
	end
	return nBonusModifier, sBonusDesc
end

--
-- Custom data info
--

function getBaseCustomData(sActorType, nodeActor, nodeSource, sTraitType, rUserData)

	local sSender = nil
	local sIcon = nil

	if nodeActor then
		local sCharType, nodeChar = asCharacterType(sActorType, nodeActor)
		if sCharType == "pc" then
			sSender, sIcon = CharacterSheetCommon.getCustomChatMessageData(nodeChar)
		elseif sCharType == "npc" then
			sSender, sIcon = NPCSheetCommon.getCustomChatMessageData(nodeChar)
		end
	end

	local rCustomData = {}
	rCustomData.sender = sSender
	rCustomData.icon = sIcon
	if nodeSource then
		rCustomData.nodename = nodeSource.getNodeName()
	end
	if sTraitType then
		rCustomData.subtypes = "[".. sTraitType .."]" .. (rCustomData.subtypes or "")
	end
	if rUserData then
		rCustomData.userdata = StringManager.convertTableToString(rUserData)
	end

	return rCustomData
end

function getTraitCustomData(sActorType, nodeActor, nodeTrait, sTraitType, rUserData)
	local rCustomData = getBaseCustomData(sActorType, nodeActor, nodeTrait, sTraitType, rUserData)
	local rActor = CharacterManager.getActorShortcut(sActorType, nodeActor)
	if hasWildDie(rActor) then
		rCustomData.wilddie = getWildDie(rActor, nodeTrait)
	end
	return rCustomData
end

function getAttackCustomData(sActorType, nodeActor, nodeAttack, sTraitType, rUserData)
	local rCustomData = getBaseCustomData(sActorType, nodeActor, nodeAttack, sTraitType, rUserData)
	local rActor = CharacterManager.getActorShortcut(sActorType, nodeActor)
	if hasWildDie(rActor) then
		local _, nodeChar = asCharacterType(rActor)
		local nodeTrait = SkillManager.getSkillNode(nodeChar, getAssociatedSkill(nodeChar, nodeAttack))
		rCustomData.wilddie = getWildDie(rActor, nodeTrait, nodeAttack)
	end
	rCustomData.fumble = getFumble(nodeAttack)
	rCustomData.isranged = getRanged(nodeAttack)
	return rCustomData
end

function getDamageCustomData(sActorType, nodeActor, nodeAttack, sTraitType, rUserData)
	local rCustomData = getBaseCustomData(sActorType, nodeActor, nodeAttack, sTraitType, rUserData)
	rCustomData.AP = getArmorPiercing(nodeAttack)
	if AbilityManager.hasAbilityOrEffect(sActorType, nodeActor, "Right Between the Eyes") then
	--rCustomData.bonusdamage = getBonusDamage(nodeAttack)
		local rWeapon = getAttackProperties(nodeAttack) or {}
		if rWeapon.isranged or isArcaneAttack(nodeAttack) then
			rCustomData.bonusdamage = DB.getValue(nodeAttack, "bonusdamage", "d8")
		else
			rCustomData.bonusdamage = getBonusDamage(nodeAttack)
		end
	else
		rCustomData.bonusdamage = getBonusDamage(nodeAttack)
	end
	rCustomData.nonlethal = getNonLethal(nodeAttack)
	return rCustomData
end

function getFumble(nodeAttack)
	if not nodeAttack then
		return 0
	else
		return DB.getValue(nodeAttack, "fumble", 0)
	end
end

function getRanged(nodeAttack)
	local rWeapon = getAttackProperties(nodeAttack) or {}
	if rWeapon.isranged or isArcaneAttack(nodeAttack) then
		return 1
	else
		return 0
	end
end

function getNonLethal(nodeAttack)
	if nodeAttack and DB.getValue(nodeAttack, "damagetype", "") == "nonlethal" then
		return 1
	else
		return 0
	end
end

function getArmorPiercing(nodeAttack)
	if not nodeAttack then
		return 0
	else
		return DB.getValue(nodeAttack, "armorpiercing", 0)
	end
end

function getBonusDamage(nodeAttack)
	if not nodeAttack then
		return "d6"
	else
		return DB.getValue(nodeAttack, "bonusdamage", "d6")
	end
end

function getAssociatedSkill(nodeChar, nodeAttack)
	if not nodeChar or not nodeAttack then
		return ""
	else
		if isArcaneAttack(nodeAttack) then
			return getPowerProperty(nodeChar, "skill") or DB.getValue(nodeAttack, "name", "")
		else
			return getAttackProperty(nodeAttack, "associatedskill") or ""
		end
	end
end

--
-- Attack properties
--

function getAttackProperties(nodeAttack)
	if nodeAttack then
		return WeaponPropertyManager.getWeaponTypeByTitle(DB.getValue(nodeAttack, "traittype"))
	end
end

function getAttackProperty(nodeAttack, sProperty)
	local rWeapon = getAttackProperties(nodeAttack)
	return rWeapon and rWeapon[sProperty]
end

--
-- Power properties
--

function getPowerProperties(nodeChar)
	if nodeChar then
		return ArcanePropertyManager.getArcaneConfig(DB.getValue(nodeChar, "arcanetype", ""))
	end
end

function getPowerProperty(nodeChar, sProperty)
	local rPower = getPowerProperties(nodeChar)
	return rPower and rPower[sProperty]
end

--
-- Helpers
--

function isArcaneAttack(nodeAttack)
	if nodeAttack then
		return StringManager.equals(DB.getValue(nodeAttack, "traittype"), ArcanePropertyManager.ARCANE_TYPE)
	end
end

function isAttributeNode(nodeTrait)
	if nodeTrait then
		local sTarget = nodeTrait.getName()
		for _, sAttribute in pairs(Global.AttributeNodes) do
			if sTarget == sAttribute then
				return true
			end
		end
	end
end

function asCharacterType(vActor, vSource)
	if type(vActor) == "table" then
		vSource = DB.findNode(vActor.recordname)
		vActor = vActor.class
	end
	if type(vActor) == "string" and vSource then
		if vSource and vActor == "ct" then
			local _, sRecord = DB.getValue(vSource, "link")
			if sRecord then
				vSource = DB.findNode(sRecord)
				if NPCSheetCommon.isNPC(vSource) then
					vActor = "npc"
				else
					vActor = "pc"
				end
			end
		end
	end
	return vActor, vSource
end

function getTraitNode(sActorType, nodeActor, sNodeName)
	local _, nodeChar = asCharacterType(sActorType, nodeActor)
	if nodeChar then
		return nodeChar.getChild(sNodeName)
	end
end

function getEffectAdjustedAttribute(sActorType, nodeActor, sAttribute)
	local _, nodeChar = asCharacterType(sActorType, nodeActor)
	if not nodeChar then
		return { "d4" }, 0
	end
	local nodeAttribute = DB.createChild(nodeChar, sAttribute, "dice")
	local aAttributeDice = nodeAttribute.getValue() or {}
	local nAttributeModifier = DB.getValue(nodeChar, sAttribute .. "Mod", 0)
	local rTraitEffect = EffectManager.getTraitAdjustmentEffect(sActorType, nodeActor, nodeAttribute)
	local sTraitEffects = nil
	if rTraitEffect then
		aAttributeDice, nAttributeModifier = SWDiceManager.getAdjustedDice(nodeAttribute, rTraitEffect)
		sTraitEffects = rTraitEffect.effectString
	end
	return aAttributeDice, nAttributeModifier, sTraitEffects
end
